class NoModificationMeta(type):
    
    def __setattr__(self, name, value):
        # don't allow changing the value of any attribute
        if hasattr(self, name):
            raise ValueError(f"Cannot reassign {name}")
        super().__setattr__(name, value)
    
    def __delattr__(self, name):
        # don't allow deleting any attribute
        raise ValueError(f"Cannot delete {name}")


class BaseEnum(metaclass=NoModificationMeta):
    """Inherit this class for creating enum in the system"""


class EmailStatus(BaseEnum):
    PENDING = 0
    SUCCESS = 1
    ERROR = 2



import os
from easy_email.enums import EmailStatus
from easy_email.validators import TemplateNameValidator
from easy_email.settings import settings
from django.db import models


def upload_to(instance, filename):
    file_path = os.path.join(settings.EASY_EMAIL_ATTACHMENT_UPLOAD_PATH, filename)
    return file_path

def storage_backend():
    return settings.EASY_EMAIL_ATTACHMENT_STORAGE_BACKEND


class Attachment(models.Model):
    file = models.FileField(upload_to=upload_to, max_length=500, null=True, storage=storage_backend)
    created_at = models.DateTimeField(auto_now_add=True, null=True, blank=True)
    updated_at = models.DateTimeField(auto_now=True, null=True, blank=True)
    
    def __str__(self):
        return self.file.name if self.file else f"File ID - {self.id}"


class Email(models.Model):
    """
    Represents an email triggered from the system
    """

    EMAIL_STATUS = [
        [EmailStatus.PENDING, "Pending"],
        [EmailStatus.SUCCESS, "Success"],
        [EmailStatus.ERROR, "Error"],
    ]

    subject = models.CharField(max_length=500, null=True, blank=True)
    body = models.TextField(null=True, blank=True)
    recipients = models.TextField(null=True, blank=True)
    from_email = models.CharField(max_length=100, null=True, blank=True)
    cc = models.TextField(null=True, blank=True)
    attachments = models.ManyToManyField(Attachment, blank=True)
    send_time = models.DateTimeField(null=True, blank=True)
    status = models.IntegerField(choices=EMAIL_STATUS, default=EmailStatus.PENDING, null=True, blank=True)
    logs = models.TextField(null=True, blank=True)
    created_at = models.DateTimeField(auto_now_add=True, null=True, blank=True)
    updated_at = models.DateTimeField(auto_now=True, null=True, blank=True)

    def __str__(self):
        return self.subject or f"Email ID - {self.id}"


class Template(models.Model):
    name = models.CharField(max_length=50, validators=[TemplateNameValidator()], unique=True,
        help_text="Only underscore (_), lowercase characters, and numbers are allowed. Name cannot start with a number.")
    content = models.TextField(null=True, blank=True)
    created_at = models.DateTimeField(auto_now_add=True, null=True, blank=True)
    updated_at = models.DateTimeField(auto_now=True, null=True, blank=True)

    def __str__(self):
        return self.name or f"Template ID - {self.id}"



import filetype
import traceback
from typing import Optional, Union, List
from django.utils import timezone
from django.conf import settings
from django.core.mail import EmailMultiAlternatives, get_connection
from easy_email.enums import EmailStatus
from easy_email.exceptions import InvalidFileFormat
from easy_email.models import Attachment, Email


class BaseEmailProcessor:
    def __init__(self, subject, email_body, recipient_list,
            from_email=None, fail_silently=True, **kwargs):
        """
        Initializes the email processor with the given email details and optional parameters.

        Args:
        - subject (str): The subject of the email.
        - email_body (str): The body content of the email (plain text).
        - recipient_list (list): A list of email addresses to which the email will be sent.
        - from_email (str, optional): The sender's email address. If not provided, a default will be used.
        - fail_silently (bool, optional): A flag that determines whether to suppress exceptions if an error occurs during email sending. Default is True (suppress errors).
        
        Keyword Arguments (kwargs):
        - files (list, optional): A list of `Attachment` objects that will be included with the email. The list is passed via `kwargs`.

        Attributes:
        - subject (str): The email subject.
        - recipient_list (list): The list of recipients for the email.
        - from_email (str): The sender's email address.
        - fail_silently (bool): Whether to fail silently or raise errors when sending the email.
        - message (str): The plain text body of the email.
        - html_message (str): The HTML body of the email (same as `message` in this case).
        - send_time (datetime, optional): The time at which the email should be sent. Defaults to None (send immediately).
        """
        self.subject = subject
        self.recipient_list = recipient_list
        self.from_email = from_email
        self.fail_silently = fail_silently
        self.message = email_body
        self.html_message = email_body
        self.send_time = None
        self.kwargs = kwargs
    
    def get_connection(self):
        connection_data = self.get_connection_data()
        return get_connection(**connection_data)
    
    def get_send_time(self, raw_send_time=None):
        raise NotImplementedError("subclass must implement this method: `get_send_time`")
    
    def send(self, send_time: Optional[Union[None, timezone.datetime, int]]=None):
        """
        Arguments:
        - send_time (Optional[Union[None, datetime, int]]): Determines when to send the email.
            - None: Sends the email instantly.
            - datetime: Sends the email at the specified future datetime.
            - int: Sends the email after the specified number of seconds (delay).
            
        If the datetime is in the past, the function will raise an error. If an integer is provided, it will be interpreted as a delay in seconds, and the email will be sent after that duration.
        """
        send_time = self.get_send_time(send_time)
        self._send(send_time)

    def _process_files(self) -> List[Attachment]:
        """
        This method processes the 'files' parameter from kwargs, ensuring that all
        files provided are instances of the Attachment class. It returns a list of valid 
        Attachment objects after performing the check.

        Input:
        - No explicit input, uses the `files` keyword argument passed to the class.

        Output:
        - List of `Attachment` objects (valid files). If any of the files are not 
        `Attachment` objects, an exception is raised.

        Raises:
        - InvalidFileFormat: If any item in the `files` list is not an instance of the 
        `Attachment` class.
        """
        # Extract the "files" from kwargs, default to an empty list if not provided
        attachments = self.kwargs.pop("files", []) or []

        # Check that all items in attachments are instances of Attachment
        for attachment in attachments:
            if not isinstance(attachment, Attachment):
                # Raise an error if the file is not an instance of the Attachment class
                raise InvalidFileFormat("All files must be instances of the Attachment class")
        
        # Return the list of valid Attachment objects
        return attachments

    def _send(self, send_time):
        connection = self.get_connection()

        attachments = self._process_files()
        email_obj = self._save_email(send_time)
        
        try:
            msg = EmailMultiAlternatives(
                subject=self.subject,
                body=self.message,
                from_email=self.from_email or self.get_default_from_email(),
                to=self.recipient_list,
                connection=connection,
                cc=self.kwargs.get('cc'),
            )
            msg.attach_alternative(self.html_message, "text/html")
            # attach the files.
            for attachment in attachments:
                msg.attach(
                    filename=attachment.file.name,
                    content=attachment.file.read(),
                    mimetype=filetype.guess_mime(attachment.file.read()),
                )
            msg.send(self.fail_silently)

            # update email status
            email_obj.status = EmailStatus.SUCCESS
            email_obj.save()
        except Exception as e:
            email_obj.status = EmailStatus.ERROR
            email_obj.logs = "".join(traceback.format_exception(None, e, e.__traceback__))
            email_obj.save()

    def get_connection_data(self):
        connection_data = {
            "host": settings.EMAIL_HOST,
            "port": settings.EMAIL_PORT,
            "username": settings.EMAIL_HOST_USER,
            "password": settings.EMAIL_HOST_PASSWORD,
            "use_tls": settings.EMAIL_USE_TLS,
        }
        return connection_data

    def get_default_from_email(self):
        return settings.DEFAULT_FROM_EMAIL

    def _save_email(self, send_time):
        attachment_objs = self._process_files()
        email_obj = Email.objects.create(
            subject=self.subject,
            body=self.message,
            recipients=self.recipient_list,
            from_email=self.from_email or self.get_default_from_email(),
            cc=self.kwargs.get('cc'),
        )
        
        email_obj.send_time = send_time
        email_obj.attachments.set(attachment_objs)
        email_obj.save()
        return email_obj




# Generated by Django 3.2 on 2025-04-16 21:06

from django.db import migrations, models
import easy_email.models
import easy_email.validators


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('file', models.FileField(max_length=500, null=True, upload_to=easy_email.models.upload_to)),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True)),
                ('updated_at', models.DateTimeField(auto_now=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Template',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Only underscore (_), lowercase characters, and numbers are allowed. Name cannot start with a number.', max_length=50, unique=True, validators=[easy_email.validators.TemplateNameValidator()])),
                ('content', models.TextField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True)),
                ('updated_at', models.DateTimeField(auto_now=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Email',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('subject', models.CharField(blank=True, max_length=500, null=True)),
                ('body', models.TextField(blank=True, null=True)),
                ('recipients', models.TextField(blank=True, null=True)),
                ('from_email', models.CharField(blank=True, max_length=100, null=True)),
                ('cc', models.TextField(blank=True, null=True)),
                ('send_time', models.DateTimeField(blank=True, null=True)),
                ('status', models.IntegerField(blank=True, choices=[[0, 'Pending'], [1, 'Success'], [2, 'Error']], default=0, null=True)),
                ('logs', models.TextField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True)),
                ('updated_at', models.DateTimeField(auto_now=True, null=True)),
                ('attachments', models.ManyToManyField(blank=True, to='easy_email.Attachment')),
            ],
        ),
    ]



from django.template import TemplateDoesNotExist
from django.template.loaders.base import Loader as BaseLoader
from django.template.base import Origin
from easy_email.models import Template


class DatabaseTemplateLoader(BaseLoader):

    def get_template_sources(self, template_name):
        origin = Origin(template_name, template_name=template_name)
        return [origin]

    def get_contents(self, template_name):
        try:
            # Try to get the template content from the database
            template = Template.objects.get(name=template_name)
            return template.content
        except Template.DoesNotExist:
            raise TemplateDoesNotExist(template_name)




from django.template import Template as TemplateEngine, Context
from easy_email.exceptions import TemplateNotFound
from easy_email.models import Template


def render_email_template(template_name, context=None, request=None):
    try:
        template = Template.objects.get(name=template_name)
    except:
        raise TemplateNotFound(template_name)
    
    template_engine = TemplateEngine(template_string=template.content)
    context = Context({'request': request, **context })
    email_content = template_engine.render(context)
    return email_content





from easy_email.exceptions import InvalidEmailProcessor
from easy_email.processors.base import BaseEmailProcessor
from django.utils import timezone


class DefaultEmailProcessor(BaseEmailProcessor):
    
    def get_send_time(self, raw_send_time=None):
        if raw_send_time:
            raise InvalidEmailProcessor("`DefaultEmailProcessor` doesn't support email scheduling, "
                "please use `CeleryEmailProcessor` instead.")
        return timezone.now()




import re
from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible


@deconstructible
class TemplateNameValidator:
    """
    Validator for Template model's name field.
    Ensures that the name:
    - Contains only lowercase letters, numbers, and underscores.
    - Does not start with a number.
    """
    message = (
        "Name can only contain underscores (_), lowercase letters, and numbers, "
        "and cannot start with a number."
    )
    pattern = r'^[a-z_][a-z0-9_]*$'
    code = 'invalid'

    def __init__(self, message=None, code=None):
        if message is not None:
            self.message = message
        if code is not None:
            self.code = code

    def __call__(self, value):
        if not re.fullmatch(self.pattern, value):
            raise ValidationError(self.message, code=self.code)



from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string
from django.core.files.storage import default_storage


class EasyEmailSettings:
    # settings variables
    EASY_EMAIL_ATTACHMENT_UPLOAD_PATH = None
    EASY_EMAIL_ATTACHMENT_STORAGE_BACKEND = None

    # set _instance to None initially
    _instance = None

    def __new__(cls, *args, **kwargs):
        # follow singleton design pattern, so that only one setting instance is created
        if not cls._instance:
            cls._instance = super(EasyEmailSettings, cls).__new__(cls, *args, **kwargs)
            cls._instance._initialize_settings()
        return cls._instance

    def _initialize_settings(self):
        self._validate_and_set_attachments_path()
        self._validate_and_set_storage_backend()

    def _validate_and_set_attachments_path(self):
        self.EASY_EMAIL_ATTACHMENT_UPLOAD_PATH = getattr(django_settings, 'EASY_EMAIL_ATTACHMENT_UPLOAD_PATH', 'easy_email/attachments')
        if type(self.EASY_EMAIL_ATTACHMENT_UPLOAD_PATH) != str:
            raise ImproperlyConfigured("EASY_EMAIL_ATTACHMENT_UPLOAD_PATH must be str type")

    def _validate_and_set_storage_backend(self):
        storage_backend = getattr(django_settings, 'EASY_EMAIL_ATTACHMENT_STORAGE_BACKEND', None)
        if storage_backend:
            if type(storage_backend) != str:
                raise ImproperlyConfigured("EASY_EMAIL_ATTACHMENT_STORAGE_BACKEND must be str type")
            # Attempt to import the storage backend if it's provided
            try:
                self.EASY_EMAIL_ATTACHMENT_STORAGE_BACKEND = import_string(storage_backend)()
            except (ImportError, AttributeError) as e:
                raise ImproperlyConfigured(f"Could not import storage backend '{storage_backend}': {e}")
        else:
            self.EASY_EMAIL_ATTACHMENT_STORAGE_BACKEND = default_storage

settings = EasyEmailSettings()



import pathlib
from setuptools import setup

# The directory containing this file
BASE_PATH = pathlib.Path(__file__).resolve().parent

# The text of the README file
README = (BASE_PATH / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="django-easy-email",
    version="1.0.0",
    description="A Django extension designed to streamline email management with powerful templates, scheduling, storage, and more.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/farhad0085/django-easy-email",
    author="Farhad Hossain",
    author_email="farhadhossain0085@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
    packages=[
        "easy_email",
        "easy_email/migrations",
        "easy_email/loaders",
        "easy_email/backends",
        "easy_email/processors",
    ],
    include_package_data=True,
    install_requires=[
        "django>=3.2",
        "filetype>=1.2.0",
    ],
)

# build
# python setup.py sdist bdist_wheel
# upload to pypi
# twine upload dist/*


from django.contrib import admin
from easy_email.models import Attachment, Email, Template


class AttachmentAdmin(admin.ModelAdmin):
    list_display = ['id', 'file', 'created_at', 'updated_at']


class EmailAdmin(admin.ModelAdmin):
    list_display = [
        "id", "subject", "recipients", "send_time",
        "status", "created_at", "updated_at"
    ]
    search_fields = ["subject", "recipients"]
    list_display_links = ["id", "subject"]

    def has_add_permission(self, request) -> bool:
        return False


class TemplateAdmin(admin.ModelAdmin):
    list_display = ['id', 'name', 'created_at', 'updated_at']
    search_fields = ['name']
    list_display_links = ['id', 'name']


admin.site.register(Email, EmailAdmin)
admin.site.register(Attachment, AttachmentAdmin)
admin.site.register(Template, TemplateAdmin)



import filetype
import traceback
from datetime import timedelta
from django.utils import timezone
from django.core.mail import EmailMultiAlternatives, get_connection
from easy_email.enums import EmailStatus
from easy_email.exceptions import InvalidEmailProcessor, InvalidSendTime
from easy_email.processors.base import BaseEmailProcessor
from easy_email.models import Attachment, Email
try:
    from celery import current_app
except ImportError as exc:
    raise ImportError(
            "To use CeleryEmailProcessor, you must've installed and setup celery!"
        ) from exc


@current_app.task
def schedule_email(subject, message, from_email, recipient_list, cc, fail_silently, email_id, connection_data, attachment_ids):
    email_obj = Email.objects.get(id=email_id)
    attachments = Attachment.objects.filter(id__in=attachment_ids)
    connection = get_connection(**connection_data)
    
    try:
        msg = EmailMultiAlternatives(
            subject=subject,
            body=message,
            from_email=from_email,
            to=recipient_list,
            connection=connection,
            cc=cc,
        )
        # attach html text
        msg.attach_alternative(message, "text/html")
        # attach the files.
        for attachment in attachments:
            msg.attach(
                filename=attachment.file.name,
                content=attachment.file.read(),
                mimetype=filetype.guess_mime(attachment.file.read()),
            )
        msg.send(fail_silently)

        # update email status
        email_obj.status = EmailStatus.SUCCESS
        email_obj.save()
    except Exception as e:
        email_obj.status = EmailStatus.ERROR
        email_obj.logs = "".join(traceback.format_exception(None, e, e.__traceback__))
        email_obj.save()


class CeleryEmailProcessor(BaseEmailProcessor):
    
    def get_send_time(self, raw_send_time=None):
        if not raw_send_time:
            raise InvalidEmailProcessor("If you don't want to schedule the email, "
                "please use `DefaultEmailProcessor` instead.")
        
        current_time = timezone.now()  # Get the current time with timezone info
        
        # If raw_send_time is a datetime, check if it's in the future
        if isinstance(raw_send_time, timezone.datetime):
            if raw_send_time > current_time:
                # If the send time is in the future, process the email
                return raw_send_time
            else:
                # Raise an exception if the send time is in the past
                raise InvalidSendTime(f"The specified datetime '{raw_send_time}' is in the past")
        
        # If raw_send_time is an integer (in seconds), schedule the email after that delay
        elif isinstance(raw_send_time, int):
            # Schedule the email by adding seconds to the current time
            return current_time + timedelta(seconds=raw_send_time)

        else:
            # Raise an exception for invalid send_time type
            raise InvalidSendTime()

    def send(self, send_time):
        send_time = self.get_send_time(send_time)
        email_obj = self._save_email(send_time)
        # set email status to Pending initially
        email_obj.status = EmailStatus.PENDING
        email_obj.save()
        connection_data = self.get_connection_data()

        # no way to call a class method using .apply_async() or .delay()
        # so creating a function outside the class to handle the scheduling
        schedule_email.apply_async((
            self.subject,
            self.message,
            self.from_email or self.get_default_from_email(),
            self.recipient_list,
            self.kwargs.get('cc'),
            self.fail_silently,
            email_obj.id,
            connection_data,
            [att.id for att in self._process_files()]
        ), eta=send_time)



class BaseException(Exception):
    """The base exception used in the system."""
    
    def __init__(self, msg=None):
        # Default message if none is provided
        if msg is None:
            msg = "An error occurred in the system"
        super().__init__(msg)


class TemplateNotFound(BaseException):
    """The exception raised when a template is not found."""
    
    def __init__(self, template_name=None):
        msg = f"Template '{template_name}' doesn't exist" if template_name else "Template doesn't exist"
        super().__init__(msg)


class InvalidSendTime(BaseException):
    """Raised when the specified send time is in the past."""
    
    def __init__(self, msg=None):
        if not msg:
            msg = "The specified send time is invalid"
        super().__init__(msg)


class InvalidFileFormat(BaseException):
    """Raised when the specified send time is in the past."""
    
    def __init__(self, msg=None):
        if not msg:
            msg = "Invalid file format"
        super().__init__(msg)


class InvalidEmailProcessor(BaseException):
    """raises when user use a wrong email processor"""
